@@ -58,6 +58,11 @@
 #include "avctp.h"
 #include "avrcp.h"
 
+/* AV/C Panel 1.23, page 76:
+ * command with the pressed value is valid for two seconds
+ */
+#define AVC_PRESS_TIMEOUT	2
+
 #define QUIRK_NO_RELEASE 1 << 0
 
 /* Message types */
@@ -114,7 +119,7 @@ struct avc_header {
 
 struct avctp_state_callback {
 	avctp_state_cb cb;
-	void *user_data;
+	struct audio_device *dev;
 	unsigned int id;
 };
 
@@ -192,8 +197,17 @@ struct avctp {
 	struct avctp_channel *control;
 	struct avctp_channel *browsing;
 
+	struct avctp_passthrough_handler *handler;
+
 	uint8_t key_quirks[256];
-	struct key_pressed *key;
+	struct key_pressed key;
+	bool initiator;
+};
+
+struct avctp_passthrough_handler {
+	avctp_passthrough_cb cb;
+	void *user_data;
+	unsigned int id;
 };
 
 struct avctp_pdu_handler {
@@ -215,6 +229,20 @@ static struct {
 	uint8_t avc;
 	uint16_t uinput;
 } key_map[] = {
+	{ "SELECT",		AVC_SELECT,		KEY_SELECT },
+	{ "UP",			AVC_UP,			KEY_UP },
+	{ "DOWN",		AVC_DOWN,		KEY_DOWN },
+	{ "LEFT",		AVC_LEFT,		KEY_LEFT },
+	{ "RIGHT",		AVC_RIGHT,		KEY_RIGHT },
+	{ "ROOT MENU",		AVC_ROOT_MENU,		KEY_MENU },
+	{ "CONTENTS MENU",	AVC_CONTENTS_MENU,	KEY_PROGRAM },
+	{ "FAVORITE MENU",	AVC_FAVORITE_MENU,	KEY_FAVORITES },
+	{ "ENTER",		AVC_ENTER,		KEY_ENTER },
+	{ "CHANNEL UP",		AVC_CHANNEL_UP,		KEY_CHANNELUP },
+	{ "CHANNEL DOWN",	AVC_CHANNEL_DOWN,	KEY_CHANNELDOWN },
+	{ "INPUT SELECT",	AVC_INPUT_SELECT,	KEY_CONFIG },
+	{ "HELP",		AVC_HELP,		KEY_HELP },
+	{ "POWER",		AVC_POWER,		KEY_POWER2 },
 	{ "VOLUME UP",		AVC_VOLUME_UP,		KEY_VOLUMEUP },
 	{ "VOLUME DOWN",	AVC_VOLUME_DOWN,	KEY_VOLUMEDOWN },
 	{ "PLAY",		AVC_PLAY,		KEY_PLAYCD },
@@ -224,6 +252,10 @@ static struct {
 	{ "BACKWARD",		AVC_BACKWARD,		KEY_PREVIOUSSONG },
 	{ "REWIND",		AVC_REWIND,		KEY_REWIND },
 	{ "FAST FORWARD",	AVC_FAST_FORWARD,	KEY_FASTFORWARD },
+	{ "F1",			AVC_F1,			KEY_F1 },
+	{ "F2",			AVC_F2,			KEY_F2 },
+	{ "F3",			AVC_F3,			KEY_F3 },
+	{ "F4",			AVC_F4,			KEY_F4 },
 	{ NULL }
 };
 
@@ -257,11 +289,25 @@ static void send_key(int fd, uint16_t key, int pressed)
 	send_event(fd, EV_SYN, SYN_REPORT, 0);
 }
 
+static gboolean auto_release(gpointer user_data)
+{
+	struct avctp *session = user_data;
+
+	session->key.timer = 0;
+
+	DBG("AV/C: key press timeout");
+
+	send_key(session->uinput, session->key.op, 0);
+
+	return FALSE;
+}
+
 static size_t handle_panel_passthrough(struct avctp *session,
 					uint8_t transaction, uint8_t *code,
 					uint8_t *subunit, uint8_t *operands,
 					size_t operand_count, void *user_data)
 {
+	struct avctp_passthrough_handler *handler = session->handler;
 	const char *status;
 	int pressed, i;
 
@@ -281,6 +327,12 @@ static size_t handle_panel_passthrough(struct avctp *session,
 		pressed = 1;
 	}
 
+	if (session->key.timer == 0 && handler != NULL) {
+		if (handler->cb(session, operands[0] & 0x7F,
+						pressed, handler->user_data))
+			goto done;
+	}
+
 	for (i = 0; key_map[i].name != NULL; i++) {
 		uint8_t key_quirks;
 
@@ -303,6 +355,22 @@ static size_t handle_panel_passthrough(struct avctp *session,
 			break;
 		}
 
+		if (pressed) {
+			if (session->key.timer > 0) {
+				g_source_remove(session->key.timer);
+				send_key(session->uinput, session->key.op, 0);
+			}
+
+			session->key.op = key_map[i].uinput;
+			session->key.timer = g_timeout_add_seconds(
+							AVC_PRESS_TIMEOUT,
+							auto_release,
+							session);
+		} else if (session->key.timer > 0) {
+			g_source_remove(session->key.timer);
+			session->key.timer = 0;
+		}
+
 		send_key(session->uinput, key_map[i].uinput, pressed);
 		break;
 	}
@@ -438,11 +506,8 @@ static void avctp_disconnected(struct avctp *session)
 		session->auth_id = 0;
 	}
 
-	if (session->key != NULL) {
-		if (session->key->timer > 0)
-			g_source_remove(session->key->timer);
-		g_free(session->key);
-	}
+	if (session->key.timer > 0)
+		g_source_remove(session->key.timer);
 
 	if (session->uinput >= 0) {
 		char address[18];
@@ -477,7 +542,11 @@ static void avctp_set_state(struct avctp *session, avctp_state_t new_state)
 
 	for (l = callbacks; l != NULL; l = l->next) {
 		struct avctp_state_callback *cb = l->data;
-		cb->cb(dev, old_state, new_state, cb->user_data);
+
+		if (cb->dev && cb->dev != dev)
+			continue;
+
+		cb->cb(dev, old_state, new_state);
 	}
 
 	switch (new_state) {
@@ -1302,6 +1371,8 @@ static void avctp_confirm_cb(GIOChannel *chan, gpointer data)
 
 	if (dev->control == NULL) {
 		btd_device_add_uuid(dev->btd_dev, AVRCP_REMOTE_UUID);
+		btd_device_add_uuid(dev->btd_dev, AVRCP_TARGET_UUID);
+
 		if (dev->control == NULL)
 			goto drop;
 	}
@@ -1549,33 +1620,27 @@ static int avctp_passthrough_release(struct avctp *session, uint8_t op)
 static gboolean repeat_timeout(gpointer user_data)
 {
 	struct avctp *session = user_data;
-	struct key_pressed *key = session->key;
 
-	avctp_passthrough_release(session, key->op);
-	avctp_passthrough_press(session, key->op);
+	avctp_passthrough_release(session, session->key.op);
+	avctp_passthrough_press(session, session->key.op);
 
 	return TRUE;
 }
 
 static void release_pressed(struct avctp *session)
 {
-	struct key_pressed *key = session->key;
+	avctp_passthrough_release(session, session->key.op);
 
-	avctp_passthrough_release(session, key->op);
+	if (session->key.timer > 0)
+		g_source_remove(session->key.timer);
 
-	if (key->timer > 0)
-		g_source_remove(key->timer);
-
-	g_free(key);
-	session->key = NULL;
+	session->key.timer = 0;
 }
 
 static bool set_pressed(struct avctp *session, uint8_t op)
 {
-	struct key_pressed *key;
-
-	if (session->key != NULL) {
-		if (session->key->op == op)
+	if (session->key.timer > 0) {
+		if (session->key.op == op)
 			return TRUE;
 		release_pressed(session);
 	}
@@ -1583,11 +1648,10 @@ static bool set_pressed(struct avctp *session, uint8_t op)
 	if (op != AVC_FAST_FORWARD && op != AVC_REWIND)
 		return FALSE;
 
-	key = g_new0(struct key_pressed, 1);
-	key->op = op;
-	key->timer = g_timeout_add_seconds(2, repeat_timeout, session);
-
-	session->key = key;
+	session->key.op = op;
+	session->key.timer = g_timeout_add_seconds(AVC_PRESS_TIMEOUT,
+							repeat_timeout,
+							session);
 
 	return TRUE;
 }
@@ -1610,7 +1674,7 @@ static gboolean avctp_passthrough_rsp(struct avctp *session, uint8_t code,
 int avctp_send_passthrough(struct avctp *session, uint8_t op)
 {
 	/* Auto release if key pressed */
-	if (session->key != NULL)
+	if (session->key.timer > 0)
 		release_pressed(session);
 
 	return avctp_passthrough_press(session, op);
@@ -1639,14 +1703,14 @@ int avctp_send_vendordep_req(struct avctp *session, uint8_t code,
 						func, user_data);
 }
 
-unsigned int avctp_add_state_cb(avctp_state_cb cb, void *user_data)
+unsigned int avctp_add_state_cb(struct audio_device *dev, avctp_state_cb cb)
 {
 	struct avctp_state_callback *state_cb;
 	static unsigned int id = 0;
 
 	state_cb = g_new(struct avctp_state_callback, 1);
 	state_cb->cb = cb;
-	state_cb->user_data = user_data;
+	state_cb->dev = dev;
 	state_cb->id = ++id;
 
 	callbacks = g_slist_append(callbacks, state_cb);
@@ -1670,6 +1734,52 @@ gboolean avctp_remove_state_cb(unsigned int id)
 	return FALSE;
 }
 
+unsigned int avctp_register_passthrough_handler(struct avctp *session,
+						avctp_passthrough_cb cb,
+						void *user_data)
+{
+	struct avctp_channel *control = session->control;
+	struct avctp_passthrough_handler *handler;
+	static unsigned int id = 0;
+
+	if (control == NULL || session->handler != NULL)
+		return 0;
+
+	handler = g_new(struct avctp_passthrough_handler, 1);
+	handler->cb = cb;
+	handler->user_data = user_data;
+	handler->id = ++id;
+
+	session->handler = handler;
+
+	return handler->id;
+}
+
+bool avctp_unregister_passthrough_handler(unsigned int id)
+{
+	GSList *l;
+
+	for (l = servers; l; l = l->next) {
+		struct avctp_server *server = l->data;
+		GSList *s;
+
+		for (s = server->sessions; s; s = s->next) {
+			struct avctp *session = s->data;
+
+			if (session->handler == NULL)
+				continue;
+
+			if (session->handler->id == id) {
+				g_free(session->handler);
+				session->handler = NULL;
+				return true;
+			}
+		}
+	}
+
+	return false;
+}
+
 unsigned int avctp_register_pdu_handler(struct avctp *session, uint8_t opcode,
 						avctp_control_pdu_cb cb,
 						void *user_data)
@@ -1822,6 +1932,7 @@ struct avctp *avctp_connect(struct audio_device *device)
 	}
 
 	session->control = avctp_channel_create(session, io, NULL);
+	session->initiator = true;
 	g_io_channel_unref(io);
 
 	return session;
@@ -1874,3 +1985,8 @@ struct avctp *avctp_get(struct audio_device *device)
 {
 	return avctp_get_internal(device->btd_dev);
 }
+
+bool avctp_is_initiator(struct avctp *session)
+{
+	return session->initiator;
+}
