@@ -32,10 +32,10 @@
 #include <glib.h>
 #include <bluetooth/sdp.h>
 #include <bluetooth/sdp_lib.h>
-#include <bluetooth/uuid.h>
 
+#include "lib/uuid.h"
+#include "btio/btio.h"
 #include "adapter.h"
-#include <btio/btio.h>
 #include "sdpd.h"
 #include "log.h"
 #include "error.h"
