@@ -25,6 +25,7 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -40,7 +41,7 @@
 #include "hciemu.h"
 
 struct hciemu {
-	gint ref_count;
+	int ref_count;
 	enum btdev_type btdev_type;
 	struct bthost *host_stack;
 	struct btdev *master_dev;
@@ -49,6 +50,7 @@ struct hciemu {
 	guint master_source;
 	guint client_source;
 	GList *post_command_hooks;
+	char bdaddr_str[18];
 };
 
 struct hciemu_command_hook {
@@ -192,20 +194,12 @@ static guint create_source_btdev(int fd, struct btdev *btdev)
 static bool create_vhci(struct hciemu *hciemu)
 {
 	struct btdev *btdev;
-	uint8_t bdaddr[6];
-	const char *str;
-	int fd, i;
+	int fd;
 
 	btdev = btdev_create(hciemu->btdev_type, 0x00);
 	if (!btdev)
 		return false;
 
-	str = hciemu_get_address(hciemu);
-
-	for (i = 5; i >= 0; i--, str += 3)
-		bdaddr[i] = strtol(str, NULL, 16);
-
-	btdev_set_bdaddr(btdev, bdaddr);
 	btdev_set_command_handler(btdev, master_command_callback, hciemu);
 
 	fd = open("/dev/vhci", O_RDWR | O_NONBLOCK | O_CLOEXEC);
@@ -308,7 +302,7 @@ struct hciemu *hciemu_ref(struct hciemu *hciemu)
 	if (!hciemu)
 		return NULL;
 
-	g_atomic_int_inc(&hciemu->ref_count);
+	__sync_fetch_and_add(&hciemu->ref_count, 1);
 
 	return hciemu;
 }
@@ -318,7 +312,7 @@ void hciemu_unref(struct hciemu *hciemu)
 	if (!hciemu)
 		return;
 
-	if (g_atomic_int_dec_and_test(&hciemu->ref_count) == FALSE)
+	if (__sync_sub_and_fetch(&hciemu->ref_count, 1) > 0)
 		return;
 
 	g_list_foreach(hciemu->post_command_hooks, destroy_command_hook, NULL);
@@ -339,10 +333,15 @@ void hciemu_unref(struct hciemu *hciemu)
 
 const char *hciemu_get_address(struct hciemu *hciemu)
 {
-	if (!hciemu)
+	const uint8_t *addr;
+
+	if (!hciemu || !hciemu->master_dev)
 		return NULL;
 
-	return "00:FA:CE:1E:55:00";
+	addr = btdev_get_bdaddr(hciemu->master_dev);
+	sprintf(hciemu->bdaddr_str, "%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X",
+			addr[5], addr[4], addr[3], addr[2], addr[1], addr[0]);
+	return hciemu->bdaddr_str;
 }
 
 bool hciemu_add_master_post_command_hook(struct hciemu *hciemu,
