@@ -41,8 +41,8 @@
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/sdp.h>
 #include <bluetooth/sdp_lib.h>
-#include <bluetooth/uuid.h>
 
+#include "lib/uuid.h"
 #include "textfile.h"
 #include "glib-helper.h"
 #include "storage.h"
@@ -144,7 +144,7 @@ int read_local_name(const bdaddr_t *bdaddr, char *name)
 	return 0;
 }
 
-sdp_record_t *record_from_string(const gchar *str)
+sdp_record_t *record_from_string(const char *str)
 {
 	sdp_record_t *rec;
 	int size, i, len;
@@ -180,8 +180,10 @@ sdp_record_t *find_record_in_list(sdp_list_t *recs, const char *uuid)
 
 		/* Extract the uuid */
 		uuid_str = bt_uuid2string(svcclass->data);
-		if (!uuid_str)
+		if (!uuid_str) {
+			sdp_list_free(svcclass, free);
 			continue;
+		}
 
 		if (!strcasecmp(uuid_str, uuid)) {
 			sdp_list_free(svcclass, free);
@@ -194,20 +196,3 @@ sdp_record_t *find_record_in_list(sdp_list_t *recs, const char *uuid)
 	}
 	return NULL;
 }
-
-int read_device_pairable(const bdaddr_t *bdaddr, gboolean *mode)
-{
-	char filename[PATH_MAX + 1], *str;
-
-	create_filename(filename, PATH_MAX, bdaddr, "config");
-
-	str = textfile_get(filename, "pairable");
-	if (!str)
-		return -ENOENT;
-
-	*mode = strcmp(str, "yes") == 0 ? TRUE : FALSE;
-
-	free(str);
-
-	return 0;
-}
