@@ -339,7 +339,7 @@ static void test_condition_complete(struct test_data *data)
 		user->expected_version = 0x06; \
 		user->expected_manufacturer = 0x003f; \
 		user->expected_supported_settings = 0x00000211; \
-		user->initial_settings = 0x00000000; \
+		user->initial_settings = 0x00000200; \
 		user->unmet_conditions = 0; \
 		tester_add_full(name, data, \
 				test_pre_setup, setup, func, NULL, \
@@ -1068,6 +1068,7 @@ static const char start_discovery_le_param[] = { 0x06 };
 static const char start_discovery_bredrle_param[] = { 0x07 };
 static const char start_discovery_valid_hci[] = { 0x01, 0x01 };
 static const char start_discovery_evt[] = { 0x07, 0x01 };
+static const char start_discovery_le_evt[] = { 0x06, 0x01 };
 
 static const struct generic_data start_discovery_not_powered_test_1 = {
 	.send_opcode = MGMT_OP_START_DISCOVERY,
@@ -1096,7 +1097,7 @@ static const struct generic_data start_discovery_valid_param_test_1 = {
 	.send_len = sizeof(start_discovery_bredrle_param),
 	.expect_status = MGMT_STATUS_SUCCESS,
 	.expect_param = start_discovery_bredrle_param,
-	.expect_len = sizeof(start_discovery_bredr_param),
+	.expect_len = sizeof(start_discovery_bredrle_param),
 	.expect_hci_command = BT_HCI_CMD_LE_SET_SCAN_ENABLE,
 	.expect_hci_param = start_discovery_valid_hci,
 	.expect_hci_len = sizeof(start_discovery_valid_hci),
@@ -1105,6 +1106,21 @@ static const struct generic_data start_discovery_valid_param_test_1 = {
 	.expect_alt_ev_len = sizeof(start_discovery_evt),
 };
 
+static const struct generic_data start_discovery_valid_param_test_2 = {
+	.send_opcode = MGMT_OP_START_DISCOVERY,
+	.send_param = start_discovery_le_param,
+	.send_len = sizeof(start_discovery_le_param),
+	.expect_status = MGMT_STATUS_SUCCESS,
+	.expect_param = start_discovery_le_param,
+	.expect_len = sizeof(start_discovery_le_param),
+	.expect_hci_command = BT_HCI_CMD_LE_SET_SCAN_ENABLE,
+	.expect_hci_param = start_discovery_valid_hci,
+	.expect_hci_len = sizeof(start_discovery_valid_hci),
+	.expect_alt_ev = MGMT_EV_DISCOVERING,
+	.expect_alt_ev_param = start_discovery_le_evt,
+	.expect_alt_ev_len = sizeof(start_discovery_le_evt),
+};
+
 static const char stop_discovery_bredrle_param[] = { 0x07 };
 static const char stop_discovery_bredrle_invalid_param[] = { 0x06 };
 static const char stop_discovery_valid_hci[] = { 0x00, 0x00 };
@@ -2630,6 +2646,9 @@ int main(int argc, char *argv[])
 	test_bredrle("Start Discovery - Success 1",
 				&start_discovery_valid_param_test_1,
 				setup_le_powered, test_command_generic);
+	test_le("Start Discovery - Success 2",
+				&start_discovery_valid_param_test_2,
+				setup_powered, test_command_generic);
 
 	test_bredrle("Stop Discovery - Success 1",
 				&stop_discovery_success_test_1,
