@@ -58,8 +58,10 @@ static GDBusClient *client = NULL;
 static GSList *players = NULL;
 static GSList *transports = NULL;
 
+static gboolean option_version = FALSE;
+static gboolean option_export = FALSE;
+
 struct player {
-	char *name;
 	char *bus_name;
 	DBusConnection *conn;
 	GDBusProxy *proxy;
@@ -168,45 +170,70 @@ static void dict_append_array(DBusMessageIter *dict, const char *key, int type,
 	dbus_message_iter_close_container(dict, &entry);
 }
 
+static void append_iter(DBusMessageIter *base, DBusMessageIter *iter)
+{
+	int type;
+
+	type = dbus_message_iter_get_arg_type(iter);
+
+	if (dbus_type_is_basic(type)) {
+		const void *value;
+
+		dbus_message_iter_get_basic(iter, &value);
+		dbus_message_iter_append_basic(base, type, &value);
+	} else if (dbus_type_is_container(type)) {
+		DBusMessageIter iter_sub, base_sub;
+		char *sig;
+
+		dbus_message_iter_recurse(iter, &iter_sub);
+
+		switch (type) {
+		case DBUS_TYPE_ARRAY:
+		case DBUS_TYPE_VARIANT:
+			sig = dbus_message_iter_get_signature(&iter_sub);
+			break;
+		default:
+			sig = NULL;
+			break;
+		}
+
+		dbus_message_iter_open_container(base, type, sig, &base_sub);
+
+		if (sig != NULL)
+			dbus_free(sig);
+
+		while (dbus_message_iter_get_arg_type(&iter_sub) !=
+							DBUS_TYPE_INVALID) {
+			append_iter(&base_sub, &iter_sub);
+			dbus_message_iter_next(&iter_sub);
+		}
+
+		dbus_message_iter_close_container(base, &base_sub);
+	}
+}
+
+static void dict_append_iter(DBusMessageIter *dict, const char *key,
+						DBusMessageIter *iter)
+{
+	DBusMessageIter entry;
+
+	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
+						NULL, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	append_iter(&entry, iter);
+
+	dbus_message_iter_close_container(dict, &entry);
+}
+
 static int parse_metadata_entry(DBusMessageIter *entry, const char *key,
 						DBusMessageIter *metadata)
 {
-	DBusMessageIter var;
-	int type;
-
-	printf("metadata %s found\n", key);
-
 	if (dbus_message_iter_get_arg_type(entry) != DBUS_TYPE_VARIANT)
 		return -EINVAL;
 
-	dbus_message_iter_recurse(entry, &var);
-
-	type = dbus_message_iter_get_arg_type(&var);
-	if (type == DBUS_TYPE_ARRAY) {
-		char **values;
-		int i;
-		DBusMessageIter array;
-
-		dbus_message_iter_recurse(&var, &array);
-
-		values = dbus_malloc0(sizeof(char *) * 8);
-
-		i = 0;
-		while (dbus_message_iter_get_arg_type(&array) !=
-							DBUS_TYPE_INVALID) {
-			dbus_message_iter_get_basic(&array, &(values[i++]));
-			dbus_message_iter_next(&array);
-		}
-
-		dict_append_array(metadata, key, DBUS_TYPE_STRING, &values, i);
-		dbus_free(values);
-	} else if (dbus_type_is_basic(type)) {
-		const void *value;
-
-		dbus_message_iter_get_basic(&var, &value);
-		dict_append_entry(metadata, key, type, &value);
-	} else
-		return -EINVAL;
+	dict_append_iter(metadata, key, entry);
 
 	return 0;
 }
@@ -419,39 +446,80 @@ static char *sender2path(const char *sender)
 	return g_strdelimit(path, ":.", '_');
 }
 
+static void copy_reply(DBusPendingCall *call, void *user_data)
+{
+	DBusMessage *msg = user_data;
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+	DBusMessage *copy;
+	DBusMessageIter args, iter;
+
+	copy = dbus_message_new_method_return(msg);
+	if (copy == NULL) {
+		dbus_message_unref(reply);
+		return;
+	}
+
+	dbus_message_iter_init_append(msg, &iter);
+
+	if (!dbus_message_iter_init(reply, &args))
+		goto done;
+
+	append_iter(&iter, &args);
+
+	dbus_connection_send(sys, copy, NULL);
+
+done:
+	dbus_message_unref(copy);
+	dbus_message_unref(reply);
+}
+
 static DBusHandlerResult player_message(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	char *owner = data;
-	dbus_uint32_t serial;
-	DBusMessage *copy, *reply;
-	DBusError err;
+	DBusMessage *copy;
+	DBusMessageIter args, iter;
+	DBusPendingCall *call;
 
-	copy = dbus_message_copy(msg);
-	dbus_message_set_destination(copy, owner);
-	reply = dbus_connection_send_with_reply_and_block(session, copy, -1,
-									&err);
-	if (!reply) {
-		if (dbus_error_is_set(&err)) {
-			fprintf(stderr, "%s\n", err.message);
-			dbus_error_free(&err);
-		}
-		dbus_message_unref(copy);
+	dbus_message_iter_init(msg, &args);
+
+	copy = dbus_message_new_method_call(owner,
+					MPRIS_PLAYER_PATH,
+					dbus_message_get_interface(msg),
+					dbus_message_get_member(msg));
+	if (copy == NULL)
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-	}
 
+	dbus_message_iter_init_append(msg, &iter);
+	append_iter(&iter, &args);
+
+	if (!dbus_connection_send_with_reply(session, copy, &call, -1))
+		goto done;
+
+	dbus_message_ref(msg);
+	dbus_pending_call_set_notify(call, copy_reply, msg, NULL);
+	dbus_pending_call_unref(call);
+
+done:
 	dbus_message_unref(copy);
 
-	copy = dbus_message_copy(reply);
-	serial = dbus_message_get_serial(msg);
-	dbus_message_set_serial(copy, serial);
-
-	dbus_message_unref(copy);
-	dbus_message_unref(reply);
-
 	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
+static struct player *find_player_by_bus_name(const char *name)
+{
+	GSList *l;
+
+	for (l = players; l; l = l->next) {
+		struct player *player = l->data;
+
+		if (strcmp(player->bus_name, name) == 0)
+			return player;
+	}
+
+	return NULL;
+}
+
 static const DBusObjectPathVTable player_table = {
 	.message_function = player_message,
 };
@@ -459,15 +527,24 @@ static const DBusObjectPathVTable player_table = {
 static void add_player(DBusConnection *conn, const char *name,
 							const char *sender)
 {
-	DBusMessage *reply = get_all(conn, name);
+	DBusMessage *reply = NULL;
 	DBusMessage *msg;
 	DBusMessageIter iter, args, properties;
 	DBusError err;
 	char *path, *owner;
+	struct player *player;
 
-	if (!reply || !adapter)
+	if (!adapter)
 		return;
 
+	player = find_player_by_bus_name(name);
+	if (player == NULL) {
+		reply = get_all(conn, name);
+		if (reply == NULL)
+			return;
+		dbus_message_iter_init(reply, &args);
+	}
+
 	msg = dbus_message_new_method_call(BLUEZ_BUS_NAME,
 					g_dbus_proxy_get_path(adapter),
 					BLUEZ_MEDIA_INTERFACE,
@@ -487,19 +564,24 @@ static void add_player(DBusConnection *conn, const char *name,
 
 	dbus_message_iter_append_basic(&iter, DBUS_TYPE_OBJECT_PATH, &path);
 
-	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
-			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
-			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
-			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &properties);
-
-	dbus_message_iter_init(reply, &args);
-
-	if (parse_properties(conn, path, &args, &properties) < 0)
-		goto done;
-
-	dbus_message_iter_close_container(&iter, &properties);
-
-	dbus_message_unref(reply);
+	if (player != NULL) {
+		if (!g_dbus_get_properties(player->conn,
+						MPRIS_PLAYER_PATH,
+						MPRIS_PLAYER_INTERFACE,
+						&iter))
+			goto done;
+	} else {
+		dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
+				DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+				DBUS_TYPE_STRING_AS_STRING
+				DBUS_TYPE_VARIANT_AS_STRING
+				DBUS_DICT_ENTRY_END_CHAR_AS_STRING,
+				&properties);
+		if (parse_properties(conn, path, &args, &properties) < 0)
+			goto done;
+		dbus_message_iter_close_container(&iter, &properties);
+		dbus_message_unref(reply);
+	}
 
 	dbus_error_init(&err);
 
@@ -598,20 +680,6 @@ done:
 	return TRUE;
 }
 
-static struct player *find_player_by_bus_name(const char *name)
-{
-	GSList *l;
-
-	for (l = players; l; l = l->next) {
-		struct player *player = l->data;
-
-		if (strcmp(player->bus_name, name) == 0)
-			return player;
-	}
-
-	return NULL;
-}
-
 static gboolean name_owner_changed(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
@@ -632,7 +700,7 @@ static gboolean name_owner_changed(DBusConnection *conn,
 	if (*new == '\0') {
 		printf("player %s at %s disappear\n", name, old);
 		remove_player(conn, old);
-	} else if (find_player_by_bus_name(name) == NULL) {
+	} else if (option_export || find_player_by_bus_name(name) == NULL) {
 		printf("player %s at %s found\n", name, new);
 		add_player(conn, name, new);
 	}
@@ -770,8 +838,12 @@ static void usage(void)
 	printf("Usage:\n");
 }
 
-static struct option main_options[] = {
-	{ 0, 0, 0, 0 }
+static GOptionEntry options[] = {
+	{ "version", 'v', 0, G_OPTION_ARG_NONE, &option_version,
+				"Show version information and exit" },
+	{ "export", 'e', 0, G_OPTION_ARG_NONE, &option_export,
+				"Export remote players" },
+	{ NULL },
 };
 
 static void connect_handler(DBusConnection *connection, void *user_data)
@@ -799,7 +871,6 @@ static void player_free(void *data)
 	if (player->transport)
 		g_dbus_proxy_unref(player->transport);
 
-	g_free(player->name);
 	g_free(player->bus_name);
 	g_free(player);
 }
@@ -851,17 +922,48 @@ static void player_control(struct player *player, DBusMessage *msg,
 						p, pending_call_free);
 }
 
+static const char *status_to_playback(const char *status)
+{
+	if (strcasecmp(status, "playing") == 0)
+		return "Playing";
+	else if (strcasecmp(status, "paused") == 0)
+		return "Paused";
+	else
+		return "Stopped";
+}
+
+static const char *player_get_status(struct player *player)
+{
+	const char *status;
+	DBusMessageIter value;
+
+	if (g_dbus_proxy_get_property(player->proxy, "Status", &value)) {
+		dbus_message_iter_get_basic(&value, &status);
+		return status_to_playback(status);
+	}
+
+	if (player->transport == NULL)
+		goto done;
+
+	if (!g_dbus_proxy_get_property(player->transport, "State", &value))
+		goto done;
+
+	dbus_message_iter_get_basic(&value, &status);
+
+	if (strcasecmp(status, "active") == 0)
+		return "Playing";
+
+done:
+	return "Stopped";
+}
+
 static DBusMessage *player_toggle(DBusConnection *conn, DBusMessage *msg,
 								void *data)
 {
 	struct player *player = data;
-	DBusMessageIter value;
 	const char *status;
 
-	if (!g_dbus_proxy_get_property(player->proxy, "Status", &value))
-		return FALSE;
-
-	dbus_message_iter_get_basic(&value, &status);
+	status = player_get_status(player);
 
 	if (strcasecmp(status, "Playing") == 0)
 		player_control(player, msg, "Pause");
@@ -923,35 +1025,18 @@ static DBusMessage *player_previous(DBusConnection *conn, DBusMessage *msg,
 
 static gboolean status_exists(const GDBusPropertyTable *property, void *data)
 {
-	DBusMessageIter iter;
 	struct player *player = data;
 
-	return g_dbus_proxy_get_property(player->proxy, "Status", &iter);
-}
-
-static const char *status_to_playback(const char *status)
-{
-	if (strcasecmp(status, "playing") == 0)
-		return "Playing";
-	else if (strcasecmp(status, "paused") == 0)
-		return "Paused";
-	else
-		return "Stopped";
+	return player_get_status(player) != NULL;
 }
 
 static gboolean get_status(const GDBusPropertyTable *property,
 					DBusMessageIter *iter, void *data)
 {
 	struct player *player = data;
-	DBusMessageIter value;
 	const char *status;
 
-	if (!g_dbus_proxy_get_property(player->proxy, "Status", &value))
-		return FALSE;
-
-	dbus_message_iter_get_basic(&value, &status);
-
-	status = status_to_playback(status);
+	status = player_get_status(player);
 
 	dbus_message_iter_append_basic(iter, DBUS_TYPE_STRING, &status);
 
@@ -1046,7 +1131,7 @@ static void set_repeat(const GDBusPropertyTable *property,
 	}
 
 	g_dbus_proxy_set_property_basic(player->proxy, "Repeat",
-					DBUS_TYPE_STRING, value,
+					DBUS_TYPE_STRING, &value,
 					property_result, GUINT_TO_POINTER(id),
 					NULL);
 }
@@ -1108,7 +1193,7 @@ static void set_shuffle(const GDBusPropertyTable *property,
 	value = shuffle ? "alltracks" : "off";
 
 	g_dbus_proxy_set_property_basic(player->proxy, "Shuffle",
-					DBUS_TYPE_STRING, value,
+					DBUS_TYPE_STRING, &value,
 					property_result, GUINT_TO_POINTER(id),
 					NULL);
 }
@@ -1221,8 +1306,6 @@ static int parse_track_entry(DBusMessageIter *entry, const char *key,
 {
 	DBusMessageIter var;
 
-	printf("metadata %s found\n", key);
-
 	if (dbus_message_iter_get_arg_type(entry) != DBUS_TYPE_VARIANT)
 		return -EINVAL;
 
@@ -1355,8 +1438,24 @@ static gboolean get_name(const GDBusPropertyTable *property,
 					DBusMessageIter *iter, void *data)
 {
 	struct player *player = data;
+	DBusMessageIter var;
+	const char *alias;
+	char *name;
 
-	dbus_message_iter_append_basic(iter, DBUS_TYPE_STRING, &player->name);
+	if (!g_dbus_proxy_get_property(player->device, "Alias", &var))
+		return FALSE;
+
+	dbus_message_iter_get_basic(&var, &alias);
+
+	if (g_dbus_proxy_get_property(player->proxy, "Name", &var)) {
+		dbus_message_iter_get_basic(&var, &name);
+		name = g_strconcat(alias, " ", name, NULL);
+	} else
+		name = g_strdup(alias);
+
+	dbus_message_iter_append_basic(iter, DBUS_TYPE_STRING, &name);
+
+	g_free(name);
 
 	return TRUE;
 }
@@ -1414,7 +1513,8 @@ static void register_player(GDBusProxy *proxy)
 {
 	struct player *player;
 	DBusMessageIter iter;
-	const char *path, *name;
+	const char *path, *alias, *name;
+	char *busname;
 	GDBusProxy *device, *transport;
 
 	if (!g_dbus_proxy_get_property(proxy, "Device", &iter))
@@ -1429,14 +1529,21 @@ static void register_player(GDBusProxy *proxy)
 	if (!g_dbus_proxy_get_property(device, "Alias", &iter))
 		return;
 
-	dbus_message_iter_get_basic(&iter, &name);
+	dbus_message_iter_get_basic(&iter, &alias);
+
+	if (g_dbus_proxy_get_property(proxy, "Name", &iter)) {
+		dbus_message_iter_get_basic(&iter, &name);
+		busname = g_strconcat(alias, " ", name, NULL);
+	} else
+		busname = g_strdup(alias);
 
 	player = g_new0(struct player, 1);
-	player->name = g_strdup(name);
-	player->bus_name = mpris_busname(player->name);
+	player->bus_name = mpris_busname(busname);
 	player->proxy = g_dbus_proxy_ref(proxy);
 	player->device = device;
 
+	g_free(busname);
+
 	players = g_slist_prepend(players, player);
 
 	printf("Player %s created\n", player->bus_name);
@@ -1552,6 +1659,9 @@ static void unregister_player(struct player *player)
 {
 	players = g_slist_remove(players, player);
 
+	g_dbus_unregister_interface(player->conn, MPRIS_PLAYER_PATH,
+						MPRIS_INTERFACE);
+
 	g_dbus_unregister_interface(player->conn, MPRIS_PLAYER_PATH,
 						MPRIS_PLAYER_INTERFACE);
 }
@@ -1689,7 +1799,7 @@ static void transport_property_changed(GDBusProxy *proxy, const char *name,
 	DBusMessageIter var;
 	const char *path;
 
-	if (strcasecmp(name, "Volume") != 0)
+	if (strcasecmp(name, "Volume") != 0 && strcasecmp(name, "State") != 0)
 		return;
 
 	if (!g_dbus_proxy_get_property(proxy, "Device", &var))
@@ -1701,6 +1811,15 @@ static void transport_property_changed(GDBusProxy *proxy, const char *name,
 	if (player == NULL)
 		return;
 
+	if (strcasecmp(name, "State") == 0) {
+		if (!g_dbus_proxy_get_property(player->proxy, "Status", &var))
+			g_dbus_emit_property_changed(player->conn,
+						MPRIS_PLAYER_PATH,
+						MPRIS_PLAYER_INTERFACE,
+						"PlaybackStatus");
+		return;
+	}
+
 	g_dbus_emit_property_changed(player->conn, MPRIS_PLAYER_PATH,
 						MPRIS_PLAYER_INTERFACE,
 						name);
@@ -1723,19 +1842,28 @@ static void property_changed(GDBusProxy *proxy, const char *name,
 
 int main(int argc, char *argv[])
 {
+	GOptionContext *context;
+	GError *error = NULL;
 	guint owner_watch, properties_watch;
 	struct sigaction sa;
-	int opt;
 
-	while ((opt = getopt_long(argc, argv, "h", main_options,
-							NULL)) != EOF) {
-		switch(opt) {
-		case 'h':
-			usage();
-			exit(0);
-		default:
-			exit(1);
-		}
+	context = g_option_context_new(NULL);
+	g_option_context_add_main_entries(context, options, NULL);
+
+	if (g_option_context_parse(context, &argc, &argv, &error) == FALSE) {
+		if (error != NULL) {
+			g_printerr("%s\n", error->message);
+			g_error_free(error);
+		} else
+			g_printerr("An unknown error occurred\n");
+		exit(1);
+	}
+
+	g_option_context_free(context);
+
+	if (option_version == TRUE) {
+		usage();
+		exit(0);
 	}
 
 	main_loop = g_main_loop_new(NULL, FALSE);
